<?php

namespace Drupal\formable\Controller;

use Drupal\Core\Controller\ControllerBase;
use Drupal\formable\FormableFormInterface;
use Drupal\Core\Render\RendererInterface;
use Drupal\Core\Url;
use Symfony\Component\HttpKernel\Exception\NotFoundHttpException;

/**
 * Controller routines for formable routes.
 */
class FormableController extends ControllerBase {

  /**
   * The renderer.
   *
   * @var \Drupal\Core\Render\RendererInterface
   */
  protected $renderer;

  /**
   * Constructs a FormableController object.
   *
   * @param \Drupal\Core\Render\RendererInterface $renderer
   *   The renderer.
   */
  public function __construct(RendererInterface $renderer) {
    $this->renderer = $renderer;
  }

  /**
   * Presents the site-wide form.
   *
   * @param \Drupal\formable\FormableFormInterface $formable_form
   *   The form to use.
   *
   * @return array
   *   The form as render array as expected by
   *   \Drupal\Core\Render\RendererInterface::render().
   *
   * @throws \Symfony\Component\HttpKernel\Exception\NotFoundHttpException
   *   Exception is thrown when user tries to access non existing default
   *   form.
   */
  public function formableSitePage(?FormableFormInterface $formable_form = NULL, $language = NULL) {

    $config = $this->config('formable.settings');

    // If there are no forms, do not display the form.
    // TODO: In theory, this should never happen, so this check can be deleted
    if (empty($formable_form)) {
      throw new NotFoundHttpException();
    }

    if ($language) {
      $languageManager = \Drupal::languageManager();
      $currentLanguage = $languageManager->getCurrentLanguage()->getId();

      if ($language !== $currentLanguage) {

        $configName = 'formable.form.' . $formable_form->id();
        $translatedConfig = $languageManager->getLanguageConfigOverride($currentLanguage, $configName);
        $route = $translatedConfig->get('route');

        $routeIsTranslated = !empty($route);

        if ($routeIsTranslated) {
          throw new NotFoundHttpException();
        }
      }
    }

    $message = $this->entityTypeManager()
      ->getStorage('formable_message')
      ->create([
        'formable_form' => $formable_form->id(),
      ]);

    $form = $this->entityFormBuilder()->getForm($message);
    $form['#title'] = $formable_form->label();
    $form['#cache']['contexts'][] = 'user.permissions';
    $this->renderer->addCacheableDependency($form, $config);

    // If the form is invalid, show a warning, but still render it to be able
    // to preview it. Access control management prevents showing the form for
    // unprivileged users.
    if (!$formable_form->isValid()) {
      $this->messenger()->addWarning($this->t("The form @name is invalid because messages can't be stored and mails can't be sent. The form won't be displayed to anonymous users.", ['@name' => $formable_form->label()]));
    }

    return $form;
  }

}
