<?php

namespace Drupal\formable\Entity;

use Drupal\Core\Config\Entity\ConfigEntityBundleBase;
use Drupal\formable\FormableFormInterface;
use Drupal\Core\Url;
use Drupal\Core\Link;
use Drupal\Component\Utility\Crypt;

/**
 * Defines the form entity.
 *
 * @ConfigEntityType(
 *   id = "formable_form",
 *   label = @Translation("Form"),
 *   label_collection = @Translation("Forms"),
 *   label_singular = @Translation("form"),
 *   label_plural = @Translation("forms"),
 *   label_count = @PluralTranslation(
 *     singular = "@count form",
 *     plural = "@count forms",
 *   ),
 *   handlers = {
 *     "access" = "Drupal\formable\FormableFormAccessControlHandler",
 *     "list_builder" = "Drupal\formable\FormableFormListBuilder",
 *     "form" = {
 *       "add" = "Drupal\formable\FormableFormEditForm",
 *       "edit" = "Drupal\formable\FormableFormEditForm",
 *       "delete" = "Drupal\Core\Entity\EntityDeleteForm"
 *     },
 *     "route_provider" = {
 *       "permissions" = "Drupal\user\Entity\EntityPermissionsRouteProviderWithCheck",
 *     }
 *   },
 *   config_prefix = "form",
 *   admin_permission = "administer formable forms",
 *   bundle_of = "formable_message",
 *   entity_keys = {
 *     "id" = "id",
 *     "label" = "label"
 *   },
 *   links = {
 *     "delete-form" = "/admin/structure/formable/manage/{formable_form}/delete",
 *     "edit-form" = "/admin/structure/formable/manage/{formable_form}",
 *     "entity-permissions-form" = "/admin/structure/formable/manage/{formable_form}/permissions",
 *     "collection" = "/admin/structure/formable",
 *     "canonical" = "/admin/forms/{formable_form}",
 *   },
 *   config_export = {
 *     "id",
 *     "label",
 *     "description",
 *     "recipients",
 *     "weight",
 *     "message",
 *     "redirect",
 *     "store_messages",
 *     "route",
 *     "strip_ip_address_parts",
 *     "honeypot",
 *     "captcha",
 *     "attached_libraries",
 *     "obfuscate_messages_for_admin",
 *     "reply_to_field",
 *     "subject_pattern",
 *   }
 * )
 */
class FormableForm extends ConfigEntityBundleBase implements FormableFormInterface {

  /**
   * The form ID.
   *
   * @var string
   */
  protected $id;

  /**
   * The human-readable label of the category.
   *
   * @var string
   */
  protected $label;

  /**
   * The description of the form.
   *
   * @var string
   */
  protected $description;

  /**
   * The message displayed to user on form submission.
   *
   * @var string
   */
  protected $message = 'Your message has been sent.';

  /**
   * List of recipient email addresses.
   *
   * @var array
   */
  protected $recipients = [];

  /**
   * The path to redirect to on form submission.
   *
   * @var string
   */
  protected $redirect;

  /**
   * The weight of the category.
   *
   * @var int
   */
  protected $weight = 0;

  /**
   * Store messages
   *
   * @var boolean
   */
  protected $store_messages = true;

  /**
   * The route of the form.
   *
   * @var string
   */
  protected $route = '';

  /**
   * The number of ip address parts to be stripped from the end.
   *
   * @var int
   */
  protected $strip_ip_address_parts = 2;

  /**
   * Honeypot settings
   *
   * @var array
   */
  protected $honeypot = [
    'enabled' => true,
    'name' => 'confirm',
  ];

  /**
   * Captcha settings
   *
   * @var array
   */
  protected $captcha = [
    'enabled' => true,
    'name' => 'simple_captcha',
    'value' => 'robot',
  ];

  /**
   * Attached libraries
   *
   * @var array
   */
  protected $attached_libraries = [];

  /**
   * Obfuscate messages for admin user
   *
   * Use case: admin can debug, but can't read messages unintentionally
   *
   * @var boolean
   */
  protected $obfuscate_messages_for_admin = false;

  /**
   * The email field name of the form to configure reply-to in mails
   *
   * @var string
   */
  protected $reply_to_field = '';

  /**
   * The pattern used for the subject in mails
   *
   * @var string
   */
  protected $subject_pattern = '[@form] New message';

  /**
   * {@inheritdoc}
   */
  public function getMessage() {
    return $this->message;
  }

  /**
   * {@inheritdoc}
   */
  public function setMessage($message) {
    $this->message = $message;
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function getRecipients() {
    return $this->recipients;
  }

  /**
   * {@inheritdoc}
   */
  public function setRecipients($recipients) {
    $this->recipients = $recipients;
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function getRedirectPath() {
    return $this->redirect;
  }

  /**
   * {@inheritdoc}
   */
  public function getRedirectUrl() {
    if ($this->redirect) {
      $url = Url::fromUserInput($this->redirect);
    }
    else {
      // TODO: default to current page
      $url = Url::fromRoute('<front>');
    }
    return $url;
  }

  /**
   * {@inheritdoc}
   */
  public function setRedirectPath($redirect) {
    $this->redirect = $redirect;
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function getWeight() {
    return $this->weight;
  }

  /**
   * {@inheritdoc}
   */
  public function setWeight($weight) {
    $this->weight = $weight;
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function getStoreMessages() {
    return $this->store_messages;
  }

  /**
   * {@inheritdoc}
   */
  public function setStoreMessages($store_messages) {
    $this->store_messages = $store_messages;
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function getRoute() {
    return $this->route;
  }

  /**
   * {@inheritdoc}
   */
  public function setRoute($route) {
    $this->route = $route;
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function getRouteLink($label = null) {

    if (empty($this->route)) return null;
    if (!$this->isValid()) return null;

    $languageManager = \Drupal::languageManager();
    $currentLanguage = $languageManager->getCurrentLanguage()->getId();

    $routeName = "formable.form.{$this->id}.{$currentLanguage}";

    // Check, if route exists to prevent throwing
    // Symfony\Component\Routing\Exception\RouteNotFoundException
    // @see https://drupal.stackexchange.com/questions/222591/how-do-i-verify-a-route-exists
    // @see https://www.drupal.org/project/drupal/issues/3029545
    /* @var \Drupal\Core\Routing\RouteProviderInterface $routeProvider */
    $routeProvider = \Drupal::service('router.route_provider');
    $routeExists = !empty($routeProvider->getRoutesByNames([$routeName]));
    if (!$routeExists) {
      if ('en' === $currentLanguage) return null;
      $routeName = "formable.form.{$this->id}.en";
      $routeExists = !empty($routeProvider->getRoutesByNames([$routeName]));
      if (!$routeExists) return null;
    }

    $uri = new Url($routeName);

    // use the translated route as label instead of `$this->route`
    if (!$label) $label = '/' . $uri->getInternalPath();

    // Pass the entity data through as options, so that alter functions do not
    // need to look up this entity again.
    $uri
      ->setOption('entity_type', $this->getEntityTypeId())
      ->setOption('entity', $this);

    return new Link($label, $uri);
  }

  /**
   * {@inheritdoc}
   */
  public function getDescription() {
    return $this->description;
  }

  /**
   * {@inheritdoc}
   */
  public function setDescription($description) {
    $this->description = $description;
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function getStripIpAddressParts() {
    return $this->strip_ip_address_parts;
  }

  /**
   * {@inheritdoc}
   */
  public function setStripIpAddressParts($strip_ip_address_parts) {
    $this->strip_ip_address_parts = $strip_ip_address_parts;
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function isValid() {
    return (true === $this->store_messages) || !empty($this->recipients);
  }

  /**
   * {@inheritdoc}
   */
  public function isHoneypotEnabled() {
    if (is_array($this->honeypot) && isset($this->honeypot['enabled'])) {
      return $this->honeypot['enabled'];
    }
    return true;
  }

  /**
   * {@inheritdoc}
   */
  public function getHoneypotName($unique = false) {
    $name = 'confirm';
    if (is_array($this->honeypot) && isset($this->honeypot['name'])) {
      $name = $this->honeypot['name'];
    }
    $suffix = !$unique ? '' : '_' . Crypt::hashBase64($name);
    return $name . $suffix;
  }

  /**
   * {@inheritdoc}
   */
  public function setHoneypotSettings($honeypot) {
    $this->honeypot = $honeypot;
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function getAttachedLibraries() {
    return $this->attached_libraries;
  }

  /**
   * {@inheritdoc}
   */
  public function setAttachedLibraries($attached_libraries) {
    $this->attached_libraries = $attached_libraries;
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function isCaptchaEnabled() {
    if (is_array($this->captcha) && isset($this->captcha['enabled'])) {
      return $this->captcha['enabled'];
    }
    return true;
  }

  /**
   * {@inheritdoc}
   */
  public function getCaptchaName($unique = false) {
    $name = 'confirm';
    if (is_array($this->captcha) && isset($this->captcha['name'])) {
      $name = $this->captcha['name'];
    }
    $suffix = !$unique ? '' : '_' . Crypt::hashBase64($name);
    return $name . $suffix;
  }

  /**
   * {@inheritdoc}
   */
  public function getCaptchaValue() {
    if (is_array($this->captcha) && isset($this->captcha['value'])) {
      return $this->captcha['value'];
    }
    return null;
  }

}
