<?php

namespace Drupal\formable\Entity;

use Drupal\Core\Entity\ContentEntityBase;
use Drupal\formable\MessageInterface;
use Drupal\Core\Entity\EntityTypeInterface;
use Drupal\Core\Field\BaseFieldDefinition;

/**
 * Defines the formable message entity.
 *
 * @ContentEntityType(
 *   id = "formable_message",
 *   label = @Translation("Formable message"),
 *   label_collection = @Translation("Formable messages"),
 *   label_singular = @Translation("formable message"),
 *   label_plural = @Translation("formable messages"),
 *   label_count = @PluralTranslation(
 *     singular = "@count formable message",
 *     plural = "@count formable messages",
 *   ),
 *   bundle_label = @Translation("Form"),
 *   handlers = {
 *     "access" = "Drupal\formable\FormableMessageAccessControlHandler",
 *     "storage" = "Drupal\formable\FormableSqlContentEntityStorage",
 *     "view_builder" = "Drupal\formable\MessageViewBuilder",
 *     "list_builder" = "Drupal\formable\MessageListBuilder",
 *     "form" = {
 *       "default" = "Drupal\formable\MessageForm",
 *       "delete" = "Drupal\Core\Entity\ContentEntityDeleteForm"
 *     }
 *   },
 *   admin_permission = "administer formable forms",
 *   links = {
 *     "delete-form" = "/admin/messages/{formable_message}/delete",
 *     "collection" = "/admin/messages/{formable_form}",
 *   },
 *   entity_keys = {
 *     "id" = "id",
 *     "bundle" = "formable_form",
 *     "uuid" = "uuid",
 *     "langcode" = "langcode"
 *   },
 *   bundle_entity_type = "formable_form",
 *   field_ui_base_route = "entity.formable_form.edit_form",
 *   base_table = "formable_message",
 * )
 */
class Message extends ContentEntityBase implements MessageInterface {

  /**
   * {@inheritdoc}
   */
  public function getFormableForm() {
    return $this->get('formable_form')->entity;
  }

  /**
   * {@inheritdoc}
   */
  public function getSenderMail() {
    $fieldName = $this->getFormableForm()->get('reply_to_field');
    if (empty($fieldName)) return null;
    if (!$this->hasField($fieldName)) return null;

    $field = $this->get($fieldName);
    if (!$field) return null;

    return $field->value;
  }

  /**
   * {@inheritdoc}
   */
  public static function baseFieldDefinitions(EntityTypeInterface $entity_type) {
    /** @var \Drupal\Core\Field\BaseFieldDefinition[] $fields */
    $fields = parent::baseFieldDefinitions($entity_type);

    $fields['formable_form']->setLabel(t('Form ID'))
      ->setDescription(t('The ID of the associated form.'));

    $fields['uuid']->setDescription(t('The message UUID.'));

    $fields['langcode']->setDescription(t('The message language code.'));

    $fields['created'] = BaseFieldDefinition::create('created')
      ->setLabel(t('Created'))
      ->setDescription(t('The time that the message was created.'))
      ->setReadOnly(TRUE);

    $fields['ip_address'] = BaseFieldDefinition::create('string')
      ->setLabel(t('IP address'))
      ->setDescription(t('The IP address of the submitter.'))
      ->setDefaultValueCallback(__CLASS__ . '::getIpAddress');

    $fields['referer'] = BaseFieldDefinition::create('string')
      ->setLabel(t('Referer url'))
      ->setDescription(t('The referer url of the submitter.'))
      ->setDefaultValueCallback(__CLASS__ . '::getReferer');

    return $fields;
  }

  /**
   * Get the IP address of the sender
   *
   * If the form is configured to strip parts of the address from the end,
   * these parts are replaced with 0.
   */
  public static function getIpAddress($entity) {
    $ip = \Drupal::request()->getClientIp();
    if (!$ip) return '';

    $max = 4 - (int) $entity->getFormableForm()->getStripIpAddressParts();
    $parts = explode('.', $ip);
    $ip = '';
    foreach ($parts as $k => $v) {
      if (0 !== $k) $ip .= '.';
      $ip .= $k < $max ? $v : '0';
    }

    return $ip;
  }

  public static function getReferer() {
    return \Drupal::request()->server->get('HTTP_REFERER');
  }

  /**
   * Always pass missing form parameter to routes
   *
   * Fixes broken redirect url when deleting message:
   * 'Some mandatory parameters are missing ("formable_form") to generate a URL
   * for route "entity.formable_message.collection".
   */
  protected function urlRouteParameters($rel) {
    $parameters = parent::urlRouteParameters($rel);
    $parameters['formable_form'] = $this->bundle();
    return $parameters;
  }

}
