<?php

namespace Drupal\formable\EventSubscriber;

use Drupal\Core\Config\ConfigCrudEvent;
use Drupal\Core\Config\ConfigCollectionEvents;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;
use Drupal\Core\StringTranslation\TranslationInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;

/**
 * Rebuild routes when route is changed (translated) or deleted via
 * config_translation module.
 *
 * Both events fire after the config is written to database, but before the
 * `originalData` property is overwritten.
 *
 * @see `save` method in core/lib/Drupal/Core/Config/Config.php
 * @see https://www.drupal.org/docs/develop/creating-modules/subscribe-to-and-dispatch-events
 *
 * @package Drupal\formable\EventSubscriber
 */
class ConfigEventsSubscriber implements EventSubscriberInterface {

  use StringTranslationTrait;

  /**
   * Constructs a ConfigEventsSubscriber object.
   *
   * @param \Drupal\Core\StringTranslation\TranslationInterface $stringTranslation
   *   The string translation service.
   */
  public function __construct(TranslationInterface $stringTranslation) {
    $this->stringTranslation = $stringTranslation;
  }

  /**
   * {@inheritdoc}
   *
   * @return array
   *   The event names to listen for, and the methods that should be executed.
   */
  public static function getSubscribedEvents() {
    return [
      ConfigCollectionEvents::SAVE_IN_COLLECTION => 'rebuildRoutesOnConfigChange',
      ConfigCollectionEvents::DELETE_IN_COLLECTION => 'rebuildRoutesOnConfigChange',
    ];
  }

  /**
   * React to a config object being saved
   *
   * @param \Drupal\Core\Config\ConfigCrudEvent $event
   *   Config crud event.
   */
  public function rebuildRoutesOnConfigChange(ConfigCrudEvent $event) {

    $config = $event->getConfig();

    if (!str_starts_with($config->getName(), 'formable.form.')) return;

    $oldRoute = $config->getOriginal('route');
    $newRoute = $config->get('route');
    $routeChanged = $oldRoute !== $newRoute;

    if (!$routeChanged) return;

    // TODO: use dependency injection
    \Drupal::service('router.builder')->rebuild();
    \Drupal::messenger()->addStatus($this->t('Routes were rebuilt to reflect route or validity changes.'));
  }

}
