<?php

namespace Drupal\formable;

use Drupal\formable\Entity\FormableForm;
use Drupal\field\Entity\FieldConfig;
use Drupal\field\Entity\FieldStorageConfig;
use Drupal\Core\Entity\Entity\EntityFormDisplay;
use Drupal\Core\Entity\Entity\EntityViewDisplay;
use Drupal\Component\Utility\Random;

/**
 * Helper class to generate forms and form fields programmatically
 *
 * TODO: cleanup
 * Quick and dirty implementation. Maybe some of this logic could be handled in
 * FormableForm->postSave() instead. Before I'm not sure about the best
 * implementation, it's not worth thinking about protected properties or
 * implementing this class as a service.
 */
class FormGenerator {

  public $entityType = 'formable_message';

  public $randomGenerator;
  public $formDisplay;
  public $viewDisplay;
  public $form;

  public $weight = 0;

  // TODO: add option for contrib modules to add presets
  // TODO: add privacy/ToS checkbox field
  public $presets = [
    'default' => [
      'reply_to_field' => 'email',
      'extra' => [
        'captcha' => true,
        'preview' => false,
        'langcode' => false,
      ],
      'fields' => [
        [
          'field_name' => 'name',
          'label' => 'Name',
          'type' => 'string',
          'widget_type' => 'string_textfield',
          'view_type' => 'string',
          'required' => true,
        ],
        [
          'field_name' => 'email',
          'label' => 'Email',
          'type' => 'email',
          'widget_type' => 'email_default',
          'view_type' => 'basic_string',
          'required' => true,
        ],
        [
          'field_name' => 'subject',
          'label' => 'Subject',
          'type' => 'string',
          'widget_type' => 'string_textfield',
          'view_type' => 'string',
          'required' => true,
        ],
        [
          'field_name' => 'message',
          'label' => 'Message',
          'type' => 'string_long',
          'widget_type' => 'string_textarea',
          'view_type' => 'basic_string',
          'required' => true,
        ],
      ],
    ],
    'newsletter' => [
      'reply_to_field' => 'email',
      'extra' => [
        'captcha' => true,
        'preview' => false,
        'langcode' => false,
      ],
      'fields' => [
        [
          'field_name' => 'name',
          'label' => 'Name',
          'type' => 'string',
          'widget_type' => 'string_textfield',
          'view_type' => 'string',
          'required' => true,
        ],
        [
          'field_name' => 'email',
          'label' => 'Email',
          'type' => 'email',
          'widget_type' => 'string_textfield',
          'view_type' => 'basic_string',
          'required' => true,
        ],
      ],
    ],
  ];

  /**
   * Generate form
   */
  public function generateForm(array $options) {

    $options = $this->getMergedOptions($options);

    // TODO: throw error if id is missing
    $id = $options['id'] ?? $this->getRandomName();

    $form = $this->getForm($id, $options);

    $this->createFields($id, $options['fields']);

    // enable captcha, disable preview button and language select field
    $extraOptions = [];
    if (isset($options['extra']) && \is_array($options['extra'])) {
      $extraOptions = $options['extra'];
    }
    $this->configureExtraFields($id, $extraOptions);

    // enable fields in form
    $this->getFormDisplay($id)->save();
    // enable fields in view
    $this->getViewDisplay($id)->save();

    return $form;
  }

  /**
   * Applies default configuration for extra fields
   */
  public function configureExtraFields(string $id, array $options = []) {

    $formDisplay = $this->getFormDisplay($id);

    $captchaIsEnabled = !!($options['captcha'] ?? false);
    if ($captchaIsEnabled) {
      // enable display
      $formDisplay->setComponent('simple_captcha', [
        'region' => 'content',
        'weight' => 99,
      ]);
    }

    $showPreviewButton = !!($options['preview'] ?? false);
    if (!$showPreviewButton) {
      $formDisplay->removeComponent('preview');
    }

    $showLanguageSelect = !!($options['langcode'] ?? false);
    if (!$showLanguageSelect) {
      $formDisplay->removeComponent('langcode');
    }

  }

  /**
   * Merge options with preset
   */
  public function getMergedOptions(array $options): array {

    $fields = [];
    if (isset($options['fields']) && \is_array($options['fields'])) {
      $fields = $options['fields'];
    }
    unset($options['fields']);

    if (isset($options['preset']) && \is_string($options['preset'])) {
      $preset = $this->getPreset($options['preset']);
      unset($options['preset']);

      // Fields need a custom merge to prevent naming clashes (options can
      // overwrite preset fields, e. g. to set name field to not be required)
      $presetFields = [];
      if (isset($preset['fields']) && \is_array($preset['fields'])) {
        $presetFields = $preset['fields'];
      }
      foreach ($fields as $ok => $ov) {
        if (!isset($ov['field_name'])) continue;
        foreach ($presetFields as $pk => $pv) {
          if (!isset($pv['field_name'])) continue;
          if ($ov['field_name'] === $pv['field_name']) {
            $fields[$ok] = \array_replace_recursive($pv, $ov);
            unset($presetFields[$pk]);
          }
        }
      }
      $fields = \array_merge($fields, $presetFields);

      unset($preset['fields']);
      $options = \array_merge_recursive($preset, $options);
    }

    $options['fields'] = $fields;

    return $options;
  }

  /**
   * Create multple fields from array
   */
  public function createFields(string $id, array $fields, bool $save = false) {

    foreach ($fields as $field) {
      $field['bundle'] = $id;
      $status = $this->createField($field);
      // TODO: Do something when field creation fails
    }

    if ($save) {
      // enable fields in form
      $this->getFormDisplay($id)->save();
      // enable fields in view
      $this->getViewDisplay($id)->save();
    }

  }

  /**
   * Get existing form or create a new one if it doesn't exist
   */
  public function getForm($id, $options = []) {

    if ($this->form) return $this->form;

    $form = FormableForm::load($id);
    if (!$form) {
      $form = FormableForm::create([]);
      $form = $this->applyOptionsToForm($form, $options);
      $form->save();
    }
    $this->form = $form;
    return $this->form;
  }

  /**
   * Apply options with default settings to given form
   */
  public function applyOptionsToForm($form, array $options, bool $withDefaults = true, bool $save = false) {

    $id = $options['id'] ?? $form->id();

    $possibleOptions = [
      "id",
      "label",
      "description",
      "recipients",
      "weight",
      "message",
      "redirect",
      "store_messages",
      "route",
      "strip_ip_address_parts",
      "honeypot",
      "captcha",
      "attached_libraries",
      "obfuscate_messages_for_admin",
      "reply_to_field",
      "subject_pattern",
    ];

    // apply options
    foreach ($options as $k => $v) {
      if (\in_array($k, $possibleOptions)) {
        $form->set($k, $v);
      }
    }

    if ($withDefaults && !empty($id)) {
      if (empty($options['label'])) {
        $form->set('label', \ucfirst($id));
      }
      if (empty($options['route'])) {
        $form->set('route', '/forms/' . $id);
      }
    }

    if ($save) $form->save();

    return $form;
  }

  /**
   * Get Preset
   */
  public function getPreset(string $name): array {
    return isset($this->presets[$name]) ? $this->presets[$name] : [];
  }

  /**
   * Create field and field storage and enable it in views
   */
  public function createField(array $options, bool $enabled = true) {

    if (empty($options['bundle'])) return false;

    $bundle = $options['bundle'];

    $name = $options['field_name'] ?? $this->getRandomName();
    $type = $options['type'] ?? 'string';
    $settings = isset($options['settings']) && isset($options['settings']) ? $options['settings'] : [];

    // Fields are defined per bundle, so each form can have it's own 'name' field.
    $field = FieldConfig::loadByName($this->entityType, $bundle, $name);
    if ($field) {
      // TODO: throw Exception
      return false;
    }

    // Storage is defined per entity type. If multiple forms have a 'name'
    // field, that field can either be reused when the settings (e. g.
    // required, max length) are the same or a new field (and field storage)
    // must be created with a different name.
    $needsFieldStorage = false;
    $storage = FieldStorageConfig::loadByName($this->entityType, $name);
    if (!$storage) {
      $needsFieldStorage = true;
    }
    else {
      $canBeReused = true;
      if ($type !== $storage->get('type')) {
        $canBeReused = false;
      }
      if (!empty($settings)) {
        // TODO: compare more possible settings
        $canBeReused = false;
      }

      if (!$canBeReused) {
        // TODO: throw Exception
        return false;
      }
    }

    if ($needsFieldStorage) {
      // create storage
      FieldStorageConfig::create([
        'type' => $type,
        'entity_type' => $this->entityType,
        'field_name' => $name,
        'settings' => $settings,
      ])->save();
    }

    // create field
    FieldConfig::create([
      'entity_type' => $this->entityType,
      'bundle' => $bundle,
      'field_name' => $name,
      'label' => $options['label'] ?? $name,
      'required' => !!($options['required'] ?? false),
    ])->save();

    if ($enabled) {
      $this->enableField([
        'name' => $name,
        'weight' => $options['weight'] ?? null,
        'type' => $options['widget_type'] ?? null,
        'bundle' => $bundle,
        'view_type' => $options['view_type'] ?? null,
      ]);
    }

    return true;
  }

  /**
   * Enable field in form and view displays
   */
  public function enableField(array $options): void {

    $name = $options['name'];
    $weight = $options['weight'] ?? $this->weight++;
    $type = $options['type'] ?? 'string_textfield';
    $bundle = $options['bundle'];

    // enable form display
    $this->getFormDisplay($bundle)->setComponent($name, [
      'region' => 'content',
      'type' => $type,
      'weight' => $weight,
    ]);

    // enable view display
    $type = $options['view_type'] ?? 'string';
    $this->getViewDisplay($bundle)->setComponent($name, [
      'region' => 'content',
      'type' => $type,
      'weight' => $weight,
      'label' => 'above',
    ]);

  }

  /**
   * Get form display
   */
  public function getFormDisplay(string $bundle) {

    if ($this->formDisplay) return $this->formDisplay;

    // create form display
    $formDisplay = EntityFormDisplay::load("core.entity_form_display.{$this->entityType}.{$bundle}.default");

    if (!$formDisplay) {
      $formDisplay = EntityFormDisplay::create([
        'targetEntityType' => $this->entityType,
        'bundle' => $bundle,
        'mode' => 'default',
        'status' => TRUE,
      ]);
    }

    $this->formDisplay = $formDisplay;

    return $this->formDisplay;
  }

  /**
   * Get view display
   */
  public function getViewDisplay(string $bundle) {

    if ($this->viewDisplay) return $this->viewDisplay;

    // create form display
    $display = EntityViewDisplay::load("core.entity_view_display.{$this->entityType}.{$bundle}.default");

    if (!$display) {
      $display = EntityViewDisplay::create([
        'targetEntityType' => $this->entityType,
        'bundle' => $bundle,
        'mode' => 'default',
        'status' => TRUE,
      ]);
    }

    $this->viewDisplay = $display;

    return $this->viewDisplay;
  }

  /**
   * Gets the random generator for the utility methods.
   *
   * @return \Drupal\Component\Utility\Random
   *   The random generator
   */
  protected function getRandomGenerator() {
    if ($this->randomGenerator) return $this->randomGenerator;
    $this->randomGenerator = new Random();
    return $this->randomGenerator;
  }

  /**
   * Get random string
   */
  public function getRandomName(string $prefix = 'formable_', int $length = 8): string {
    $name = $prefix . $this->getRandomGenerator()->name($length);
    return \strtolower($name);
  }

  /**
   * Reset properties
   *
   * Useful during tests when creating multiple forms programmatically
   */
  public function reset() {
    $this->form = null;
    $this->formDisplay = null;
    $this->viewDisplay = null;
    return $this;
  }

  /**
   * Rebuild routes
   *
   * Useful when creating form programmatically (UI edit form takes care of
   * route rebuilding already)
   */
  public function rebuildRoutes() {
    \Drupal::service('router.builder')->rebuild();
  }

}
