<?php

namespace Drupal\formable;

use Drupal\Component\Utility\EmailValidatorInterface;
use Drupal\Core\DependencyInjection\ContainerInjectionInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Drupal\Core\Entity\EntityForm;
use Drupal\Core\Entity\EntityTypeInterface;
use Drupal\Core\Form\ConfigFormBaseTrait;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Path\PathValidatorInterface;
use Drupal\Core\Render\Element\PathElement;

/**
 * Base form for form edit forms.
 *
 * @internal
 */
class FormableFormEditForm extends EntityForm implements ContainerInjectionInterface {
  use ConfigFormBaseTrait;

  /**
   * The email validator.
   *
   * @var \Drupal\Component\Utility\EmailValidatorInterface
   */
  protected $emailValidator;

  /**
   * The path validator.
   *
   * @var \Drupal\Core\Path\PathValidatorInterface
   */
  protected $pathValidator;

  /**
   * Constructs a new FormableFormEditForm.
   *
   * @param \Drupal\Component\Utility\EmailValidatorInterface $email_validator
   *   The email validator.
   * @param \Drupal\Core\Path\PathValidatorInterface $path_validator
   *   The path validator service.
   */
  public function __construct(EmailValidatorInterface $email_validator, PathValidatorInterface $path_validator) {
    $this->emailValidator = $email_validator;
    $this->pathValidator = $path_validator;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('email.validator'),
      $container->get('path.validator')
    );
  }

  /**
   * {@inheritdoc}
   */
  protected function getEditableConfigNames() {
    return ['formable.settings'];
  }

  /**
   * {@inheritdoc}
   */
  public function form(array $form, FormStateInterface $form_state) {
    $form = parent::form($form, $form_state);

    $formable_form = $this->entity;

    $form['label'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Label'),
      '#maxlength' => 255,
      '#default_value' => $formable_form->label(),
      '#description' => $this->t("Example: 'website feedback' or 'product information'."),
      '#required' => TRUE,
    ];
    $form['id'] = [
      '#type' => 'machine_name',
      '#default_value' => $formable_form->id(),
      '#maxlength' => EntityTypeInterface::BUNDLE_MAX_LENGTH,
      '#machine_name' => [
        'exists' => '\Drupal\formable\Entity\FormableForm::load',
      ],
      '#disabled' => !$formable_form->isNew(),
    ];

    if ($formable_form->isNew()) {
      $form['preset'] = [
        '#type' => 'select',
        '#title' => $this->t('Preset'),
        '#description' => $this->t('Choose a preset with default settings and fields.'),
        // TODO: Should not be hard coded
        '#options' => [
          'none' => $this->t('none'),
          'default' => $this->t('Default'),
          'newsletter' => $this->t('Newsletter'),
        ],
      ];
    }

    $form['description'] = [
      '#type' => 'textarea',
      '#title' => $this->t('Description'),
      '#default_value' => $formable_form->getDescription(),
      '#description' => $this->t('Room for notes about the form.'),
    ];
    $form['recipients'] = [
      '#type' => 'textarea',
      '#title' => $this->t('Recipients'),
      '#default_value' => implode(', ', $formable_form->getRecipients()),
      '#description' => $this->t("Example: 'webmaster@example.com' or 'sales@example.com,support@example.com' . To specify multiple recipients, separate each email address with a comma."),
    ];
    $form['message'] = [
      '#type' => 'textarea',
      '#title' => $this->t('Message'),
      '#default_value' => $formable_form->getMessage(),
      '#description' => $this->t('The message to display to the user after submission of this form. Leave blank for no message.'),
    ];
    $form['redirect'] = [
      '#type' => 'path',
      '#title' => $this->t('Redirect path'),
      '#convert_path' => PathElement::CONVERT_NONE,
      '#default_value' => $formable_form->getRedirectPath(),
      '#description' => $this->t('Path to redirect the user to after submission of this form. For example, type "/about" to redirect to that page. Use a relative path with a slash in front.'),
    ];
    $form['weight'] = [
      '#type' => 'weight',
      '#title' => $this->t('Weight'),
      '#default_value' => $formable_form->getWeight(),
      '#description' => $this->t('When listing forms, those with lighter (smaller) weights get listed before forms with heavier (larger) weights. Forms with equal weights are sorted alphabetically.'),
    ];
    $form['store_messages'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Store messages'),
      '#default_value' => $formable_form->getStoreMessages(),
      '#description' => $this->t('Store sent messages in the database.'),
    ];
    $form['route'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Route'),
      '#default_value' => $formable_form->getRoute(),
      '#description' => $this->t('The form will be publicly available at the configured route.'),
    ];
    $form['strip_ip_address_parts'] = [
      '#type' => 'number',
      '#title' => $this->t('Strip parts of the stored IP address from the end.'),
      '#default_value' => $formable_form->getStripIpAddressParts(),
      '#description' => $this->t('0 strips nothing, 4 strips the whole IP address. Example: 198.51.100.42 becomes 198.51.100.0 (1 part stripped) or 198.51.0.0 (2 parts stripped).'),
    ];

    // add invisible honeypot field
    $form['honeypot'] = [
      '#type' => 'details',
      '#title' => t('Honeypot settings'),
      '#description' => $this->t('Invisible field, that adds some protection against spam bots.'),
      '#open' => TRUE,
      // automatically handle array values
      '#tree' => TRUE,

      'enabled' => [
        '#type' => 'checkbox',
        '#title' => $this->t('Enable honeypot field'),
        '#default_value' => $formable_form->isHoneypotEnabled(),
      ],
      'name' => [
        '#type' => 'textfield',
        '#title' => $this->t('Field name of honeypot field'),
        '#default_value' => $formable_form->getHoneypotName(),
        '#description' => $this->t('Is used as name attribute in the hidden input field. Tricking spam bots into clicking the checkbox can be more successful with speaking names like "confirm" or "check".'),
      ],
    ];

    // add simple captcha field
    $form['captcha'] = [
      '#type' => 'details',
      '#title' => t('Simple captcha settings'),
      '#description' => $this->t('Simple captcha field, that adds protection against spam bots.'),
      '#open' => TRUE,
      // automatically handle array values
      '#tree' => TRUE,

      'enabled' => [
        '#type' => 'checkbox',
        '#title' => $this->t('Enable captcha field'),
        '#default_value' => $formable_form->isCaptchaEnabled(),
      ],
      'name' => [
        '#type' => 'textfield',
        '#title' => $this->t('Field name of captcha field'),
        '#default_value' => $formable_form->getCaptchaName(),
        '#description' => $this->t('Changing the name attribute of the captcha field might make it harder to automate bots to fill this field.'),
      ],
      'value' => [
        '#type' => 'textfield',
        '#title' => $this->t('Expected value of captcha field'),
        '#default_value' => $formable_form->getCaptchaValue(),
        '#description' => $this->t('The expected word, that site visitors should type to pass the human vs robots test. The test itself is case insensitive, so if the expected word is "human" than "human", "Human" or "HUMAN" would all pass.'),
      ],
    ];

    $form['attached_libraries'] = [
      '#type' => 'textarea',
      '#title' => t('Attached libraries'),
      '#default_value' => implode("\r\n", $formable_form->getAttachedLibraries()),
      '#description' => $this->t("List of attached libraries. Write each library on a separate line.<br /> Example: <code>core/drupal.form</code> (restores the behaviour of the core contact module and provides e. g. protection against duplicated form submissions on double click, but pollutes the page with jQuery)."),
    ];

    $form['obfuscate_messages_for_admin'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Obfuscate messages for admin user'),
      '#default_value' => $formable_form->get('obfuscate_messages_for_admin'),
      '#description' => $this->t("Use case: An admin can view and debug the list, but the content is obfuscated to prevent reading it unintentionally."),
    ];

    // reply-to settings
    // TODO: should be select field of existing field names
    // TODO: should detect field of type email automatically
    $form['reply_to_field'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Email reply-to field'),
      '#default_value' => $formable_form->get('reply_to_field'),
      '#description' => $this->t('Enter the machine name of the email field, that should be used as reply-to addres in mails.'),
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function validateForm(array &$form, FormStateInterface $form_state) {
    parent::validateForm($form, $form_state);

    // Validate each email recipient.
    if (!empty($form_state->getValue('recipients'))) {
      $recipients = explode(',', $form_state->getValue('recipients'));

      foreach ($recipients as &$recipient) {
        $recipient = trim($recipient);
        if (!$this->emailValidator->isValid($recipient)) {
          $form_state->setErrorByName('recipients', $this->t('%recipient is an invalid email address.', ['%recipient' => $recipient]));
        }
      }
      $form_state->setValue('recipients', $recipients);
      $redirect_url = $form_state->getValue('redirect');
      if ($redirect_url && $this->pathValidator->isValid($redirect_url)) {
        if (mb_substr($redirect_url, 0, 1) !== '/') {
          $form_state->setErrorByName('redirect', $this->t('The path should start with /.'));
        }
      }
    } else {
      $form_state->setValue('recipients', []);
    }

    // TODO: better validation
    // TODO: check, if path exists already
    $route = $form_state->getValue('route');
    if ($route && is_string($route)) {

      // auto-correct accidental whitespace
      $routeTrimmed = trim($route);
      if ($route !== $routeTrimmed) {
        $route = $routeTrimmed;
        $form_state->setValue('route', $route);
      }

      if (mb_substr($route, 0, 1) !== '/') {
        $form_state->setErrorByName('route', $this->t('The path should start with /.'));
      }
      elseif (mb_substr($route, 0, 1) === '//') {
        $form_state->setErrorByName('route', $this->t('The path must not start with //.'));
      }
    }

    // Convert integers to boolean
    // If `store_messages` is not converted, the entity has invalid values
    // during the `isValid() check in the save method
    $booleanFields = [
      'store_messages',
      'obfuscate_messages_for_admin',
    ];
    foreach ($booleanFields as $name) {
      $form_state->setValue($name, (bool) $form_state->getValue($name));
    }

    // validate and convert newline separated library field to array
    if (!empty($form_state->getValue('attached_libraries'))) {
      $libraries = $form_state->getValue('attached_libraries');
      if ($libraries && is_string($libraries)) {
        // convert string to array
        $libraries = $this->splitByNewLine($libraries);

        // validation
        foreach ($libraries as &$library) {

          $library = trim($library);
          if (!preg_match('#[a-z]+[^/]*/[a-z]+[^/]*#', $library)) {
            $form_state->setErrorByName('attached_libraries', $this->t('The library "@library" must have the pattern "module/library".', ['@library' => $library]));
            continue;
          }

          [$module, $lib] = explode('/', $library, 2);
          if (('core' !== $module) && !\Drupal::moduleHandler()->moduleExists($module)) {
             $form_state->setErrorByName('attached_libraries', $this->t('The module or theme "@module" in "@library" is not installed.', ['@module' => $module, '@library' => $library]));
          }
        }
        // set converted value
        $form_state->setValue('attached_libraries', $libraries);
      }
    } else {
      $form_state->setValue('attached_libraries', []);
    }
  }

  /**
   * {@inheritdoc}
   */
  public function save(array $form, FormStateInterface $form_state) {
    $formable_form = $this->entity;

    $newRoute = $formable_form->getRoute();

    if ($formable_form->isNew()) {
      $routeChanged = !empty($newRoute);
      $validityChanged = false;
    }
    else {
      $original = \Drupal::entityTypeManager()
        ->getStorage('formable_form')
        ->loadUnchanged($formable_form->id());
      $oldRoute = $original->getRoute();

      $routeChanged = $newRoute !== $oldRoute;

      $validityChanged = $original->isValid() !== $formable_form->isValid();
    }

    $status = $formable_form->save();

    // apply preset
    $preset = $form_state->getValue('preset');
    if (\is_string($preset) && !empty(\trim($preset)) && 'none' !== $preset) {
      // Create fields from preset and configure extra fields. Using
      // `$generator->generateForm()` directly would apply default values like
      // label and route, which are configured via UI already
      $generator = new FormGenerator();
      $id = $formable_form->id();
      $options = [
        'id' => $id,
        'preset' => $preset,
      ];
      $options = $generator->getMergedOptions($options);
      if (!empty($options['fields']) && \is_array($options['fields'])) {
        $generator->form = $formable_form;
        $generator->createFields($id, $options['fields']);
      }
      $generator->configureExtraFields($id);
      $generator->getFormDisplay($id)->save();
      $generator->getViewDisplay($id)->save();
    }

    if ($routeChanged || $validityChanged) {
      // TODO: works, but using cache tags would be better
      \Drupal::service('router.builder')->rebuild();

      $this->messenger()->addStatus($this->t('Routes were rebuilt to reflect route or validity changes.'));
    }

    $edit_link = $this->entity->toLink($this->t('Edit'), 'edit-form')->toString();

    $routeLink = $formable_form->getRouteLink($formable_form->label());
    if ($routeLink) {
      $view_link = $routeLink->toString();
    } else {
      $view_link = $formable_form->toLink($formable_form->label(), 'canonical')->toString();
    }

    if ($status == SAVED_UPDATED) {
      $this->messenger()->addStatus($this->t('Form %label has been updated.', ['%label' => $view_link]));
      $this->logger('formable')->notice('Form %label has been updated.', ['%label' => $formable_form->label(), 'link' => $edit_link]);
    }
    else {
      $this->messenger()->addStatus($this->t('Form %label has been added.', ['%label' => $view_link]));
      $this->logger('formable')->notice('Form %label has been added.', ['%label' => $formable_form->label(), 'link' => $edit_link]);
    }

    $form_state->setRedirectUrl($formable_form->toUrl('collection'));
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state) {

    $form = parent::buildForm($form, $form_state);

    // Add warning message on top of form if form is invalid. The check for
    // `getUserInput()` is to prevent showing the warning message after submit
    // and redirect to the list view (while the form is valid already).
    if (!$this->entity->isNew() && !$this->entity->isValid() && empty($form_state->getUserInput())) {
      $this->messenger()->addWarning($this->t("The form @name is invalid because messages can't be stored and mails can't be sent. The form won't be displayed to anonymous users.", ['@name' => $this->entity->label()]));
    }

    return $form;
  }

  /**
   * Convert newline separated string to array
   */
  public function splitByNewLine(string $text): array {
    return preg_split('~\R+~u', $text, -1, PREG_SPLIT_NO_EMPTY);
  }

}
