<?php

namespace Drupal\formable;

use Drupal\Core\Config\Entity\ConfigEntityInterface;

/**
 * Provides an interface defining a form entity.
 */
interface FormableFormInterface extends ConfigEntityInterface {

  /**
   * Returns the message to be displayed to user.
   *
   * @return string
   *   A user message.
   */
  public function getMessage();

  /**
   * Returns list of recipient email addresses.
   *
   * @return array
   *   List of recipient email addresses.
   */
  public function getRecipients();

  /**
   * Returns the path for redirect.
   *
   * @return string
   *   The redirect path.
   */
  public function getRedirectPath();

  /**
   * Returns the \Drupal\Core\Url object for redirect path.
   *
   * Empty redirect property results a \Drupal\Core\Url object of front page.
   *
   * @return \Drupal\Core\Url
   *   The redirect \Drupal\Core\Url object.
   */
  public function getRedirectUrl();

  /**
   * Returns the weight of this category (used for sorting).
   *
   * @return int
   *   The weight of this category.
   */
  public function getWeight();

  /**
   * Sets the message to be displayed to the user.
   *
   * @param string $message
   *   The message to display after form is submitted.
   *
   * @return $this
   */
  public function setMessage($message);

  /**
   * Sets list of recipient email addresses.
   *
   * @param array $recipients
   *   The desired list of email addresses of this category.
   *
   * @return $this
   */
  public function setRecipients($recipients);

  /**
   * Sets the redirect path.
   *
   * @param string $redirect
   *   The desired path.
   *
   * @return $this
   */
  public function setRedirectPath($redirect);

  /**
   * Sets the weight.
   *
   * @param int $weight
   *   The desired weight.
   *
   * @return $this
   */

  public function setWeight($weight);

  /**
   * Returns the status of storing messages.
   */
  public function getStoreMessages();

  /**
   * Set status of storing messages.
   *
   * @param bool $store_messages
   *   The desired status.
   *
   * @return $this
   */
  public function setStoreMessages($store_messages);

  /**
   * Returns the route of the form.
   */
  public function getRoute();

  /**
   * Set the route of the form.
   *
   * @param string $route
   *   The desired route.
   *
   * @return $this
   */
  public function setRoute($route);

  /**
   * Get link object of route link
   *
   * TODO: should be link template with uri_callback instead
   *
   * @param ?string $label
   *   The desired label (defaults to route)
   *
   * @return Link|null
   *   Link object
   */
  public function getRouteLink($label);

  /**
   * Returns the description of the form.
   */
  public function getDescription();

  /**
   * Set the description of the form.
   *
   * @param string $description
   *   The desired description.
   *
   * @return $this
   */
  public function setDescription($description);

  /**
   * Returns the number of ip address parts to be stripped.
   *
   * @return int
   */
  public function getStripIpAddressParts();

  /**
   * Set the the number of ip address parts to be stripped.
   *
   * @param int $strip_ip_address_parts
   *   The desired number of ip address parts to be stripped.
   *
   * @return $this
   */
  public function setStripIpAddressParts($strip_ip_address_parts);

  /**
   * If messages aren't stored and no recipients for mail delivery exist,
   * the form is not valid.
   */
  public function isValid();

  /**
   * Returns the honeypot enabled setting of the form.
   */
  public function isHoneypotEnabled();

  /**
   * Returns the honeypot name setting of the form.
   */
  public function getHoneypotName($unique = false);

  /**
   * Set the honeypot settings of the form.
   *
   * @param array $honeypot
   *   The desired honeypot settings.
   *
   * @return $this
   */
  public function setHoneypotSettings($honeypot);

  /**
   * Returns the attached libraries of the form.
   */
  public function getAttachedLibraries();

  /**
   * Set the attached libraries of the form.
   *
   * @param array $libraries
   *   The desired libraries.
   *
   * @return $this
   */
  public function setAttachedLibraries($libraries);

  /**
   * Returns the captcha enabled setting of the form.
   */
  public function isCaptchaEnabled();

  /**
   * Returns the captcha field name setting of the form.
   */
  public function getCaptchaName($unique = false);

  /**
   * Returns the captcha field value setting of the form.
   */
  public function getCaptchaValue();

}
