<?php

namespace Drupal\formable;

use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Language\LanguageManagerInterface;
use Drupal\Core\Mail\MailManagerInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\Core\StringTranslation\TranslationInterface;
use Psr\Log\LoggerInterface;

/**
 * Provides a class for handling assembly and dispatch of formable mail messages.
 */
class MailHandler implements MailHandlerInterface {

  use StringTranslationTrait;

  /**
   * Language manager service.
   *
   * @var \Drupal\Core\Language\LanguageManagerInterface
   */
  protected $languageManager;

  /**
   * Logger service.
   *
   * @var \Drupal\Core\Logger\LoggerChannelInterface
   */
  protected $logger;

  /**
   * Mail manager service.
   *
   * @var \Drupal\Core\Mail\MailManagerInterface
   */
  protected $mailManager;

  /**
   * The user entity storage handler.
   *
   * @var \Drupal\Core\Entity\EntityStorageInterface
   */
  protected $userStorage;

  /**
   * Constructs a new \Drupal\formable\MailHandler object.
   *
   * @param \Drupal\Core\Mail\MailManagerInterface $mail_manager
   *   Mail manager service.
   * @param \Drupal\Core\Language\LanguageManagerInterface $language_manager
   *   Language manager service.
   * @param \Psr\Log\LoggerInterface $logger
   *   A logger instance.
   * @param \Drupal\Core\StringTranslation\TranslationInterface $string_translation
   *   String translation service.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   */
  public function __construct(MailManagerInterface $mail_manager, LanguageManagerInterface $language_manager, LoggerInterface $logger, TranslationInterface $string_translation, EntityTypeManagerInterface $entity_type_manager) {
    $this->languageManager = $language_manager;
    $this->mailManager = $mail_manager;
    $this->logger = $logger;
    $this->stringTranslation = $string_translation;
    $this->userStorage = $entity_type_manager->getStorage('user');
  }

  /**
   * {@inheritdoc}
   */
  public function sendMailMessages(MessageInterface $message) {
    $formable_form = $message->getFormableForm();

    // Don't send mail if recipients are empty
    // For invalid mail address the mailer service takes care of creating and
    // logging error message.
    $recipients = $formable_form->getRecipients();
    if (empty($recipients) || !is_array($recipients)) {
      return null;
    }
    $to = implode(', ', $recipients);

    $params = [];
    $langcode = $this->languageManager->getDefaultLanguage()->getId();

    // Build email parameters.
    $params['formable_message'] = $message;

    // Send to the form recipient(s), using the site's default language.
    $params['formable_form'] = $formable_form;

    // Send email to the recipient(s).
    $mailerStatus = $this->mailManager->mail(
      'formable',   // module
      'page_mail',  // key to check context during hook_mail
      $to,          // form recipient(s), e. g. admin@example.com
      $langcode,    // string, used for localized message subject and body
      $params,      // params for hook_mail
      $message->getSenderMail(), // Reply-to: alice@wonderland.com
      // optional bool $send, default: true
    );

    // the 'result' key should always exist, but it didn't during unit tests
    $mailerStatusResult = isset($mailerStatus['result']) ? $mailerStatus['result'] : null;

    // only log message if mail was sent
    if ($mailerStatusResult) {
      $senderFrom = $message->getSenderMail();
      if (empty($senderFrom)) $senderFrom = 'n/a';
      $this->logger->info('New email from @sender-from in %formable_form.', [
        '@sender-from' => $senderFrom,
        '%formable_form' => $formable_form->label(),
      ]);
    }

    return $mailerStatusResult;
  }

}
