<?php

namespace Drupal\formable;

use Drupal\Component\Datetime\TimeInterface;
use Drupal\Core\Datetime\DateFormatterInterface;
use Drupal\Core\Entity\ContentEntityForm;
use Drupal\Core\Entity\EntityRepositoryInterface;
use Drupal\Core\Entity\EntityTypeBundleInfoInterface;
use Drupal\Core\Flood\FloodInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Language\LanguageManagerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Form controller for formable message forms.
 *
 * @internal
 */
class MessageForm extends ContentEntityForm {

  /**
   * The message being used by this form.
   *
   * @var \Drupal\formable\MessageInterface
   */
  protected $entity;

  /**
   * The flood control mechanism.
   *
   * @var \Drupal\Core\Flood\FloodInterface
   */
  protected $flood;

  /**
   * The language manager service.
   *
   * @var \Drupal\Core\Language\LanguageManagerInterface
   */
  protected $languageManager;

  /**
   * The formable mail handler service.
   *
   * @var \Drupal\formable\MailHandlerInterface
   */
  protected $mailHandler;

  /**
   * The date formatter service.
   *
   * @var \Drupal\Core\Datetime\DateFormatterInterface
   */
  protected $dateFormatter;

  /**
   * Constructs a MessageForm object.
   *
   * @param \Drupal\Core\Entity\EntityRepositoryInterface $entity_repository
   *   The entity repository.
   * @param \Drupal\Core\Flood\FloodInterface $flood
   *   The flood control mechanism.
   * @param \Drupal\Core\Language\LanguageManagerInterface $language_manager
   *   The language manager service.
   * @param \Drupal\formable\MailHandlerInterface $mail_handler
   *   The formable mail handler service.
   * @param \Drupal\Core\Datetime\DateFormatterInterface $date_formatter
   *   The date service.
   * @param \Drupal\Core\Entity\EntityTypeBundleInfoInterface $entity_type_bundle_info
   *   The entity type bundle service.
   * @param \Drupal\Component\Datetime\TimeInterface $time
   *   The time service.
   */
  public function __construct(EntityRepositoryInterface $entity_repository, FloodInterface $flood, LanguageManagerInterface $language_manager, MailHandlerInterface $mail_handler, DateFormatterInterface $date_formatter, ?EntityTypeBundleInfoInterface $entity_type_bundle_info = NULL, ?TimeInterface $time = NULL) {
    parent::__construct($entity_repository, $entity_type_bundle_info, $time);
    $this->flood = $flood;
    $this->languageManager = $language_manager;
    $this->mailHandler = $mail_handler;
    $this->dateFormatter = $date_formatter;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('entity.repository'),
      $container->get('flood'),
      $container->get('language_manager'),
      $container->get('formable.mail_handler'),
      $container->get('date.formatter'),
      $container->get('entity_type.bundle.info'),
      $container->get('datetime.time')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function form(array $form, FormStateInterface $form_state) {

    $message = $this->entity;
    $formableForm = $message->getFormableForm();
    $form = parent::form($form, $form_state);
    $form['#attributes']['class'][] = 'formable-form';

    if (!empty($message->preview)) {
      $form['preview'] = [
        '#theme_wrappers' => ['container__preview'],
        '#attributes' => ['class' => ['preview']],
      ];
      $form['preview']['message'] = $this->entityTypeManager->getViewBuilder('formable_message')->view($message, 'full');
    }

    // attach configured libraries, e. g. 'core/drupal.form' to restore the
    // behaviour of the core contact module
    $attachedLibraries = $formableForm->getAttachedLibraries();
    if (is_array($attachedLibraries)) {
      foreach ($attachedLibraries as $lib) {
        $form['#attached']['library'][] = $lib;
      }
    }

    // add hidden honeypot field
    if ($formableForm->isHoneypotEnabled()) {
      $honeypotName = $formableForm->getHoneypotName();
      // prevent clashes with existing fields
      if ($message->hasField($honeypotName) || $message->hasField("field_{$honeypotName}")) {
        $honeypotName = $formableForm->getHoneypotName(true);
      }
      $form[$honeypotName] = [
        '#type' => 'checkbox',
        '#title' => $this->t("Don't click this checkbox if you are a human."),
        '#title_display' => 'after',
        '#attributes' => [
          'style' => ['display: none !important;'],
          'tabindex' => '-1',
          'autocomplete' => 'off',
        ],
        '#label_attributes' => [
          'style' => ['display: none !important;'],
        ],
      ];
    }

    // add simple captcha field
    if ($formableForm->isCaptchaEnabled()) {
      $captchaName = $formableForm->getCaptchaName();
      // prevent clashes with existing fields
      if ($message->hasField($captchaName)) {
        $captchaName = $formableForm->getCaptchaName(true);
      }
      $captchaValue = $formableForm->getCaptchaValue();
      $form[$captchaName] = [
        '#type' => 'textfield',
        '#title' => $this->t("Spam protection"),
        '#description' => $this->t("Please enter the word <strong>:word</strong>.", [
          ':word' => $captchaValue,
        ]),
        '#required' => TRUE,
      ];
    }

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function actions(array $form, FormStateInterface $form_state) {
    $elements = parent::actions($form, $form_state);
    $elements['submit']['#value'] = $this->t('Send message');
    $elements['preview'] = [
      '#type' => 'submit',
      '#value' => $this->t('Preview'),
      '#submit' => ['::submitForm', '::preview'],
      '#access' => !empty($form_state->getStorage()['form_display']->getComponent('preview')),
    ];
    return $elements;
  }

  /**
   * Form submission handler for the 'preview' action.
   */
  public function preview(array $form, FormStateInterface $form_state) {
    $message = $this->entity;
    $message->preview = TRUE;
    $form_state->setRebuild();
  }

  /**
   * {@inheritdoc}
   */
  public function validateForm(array &$form, FormStateInterface $form_state) {
    $message = parent::validateForm($form, $form_state);

    // Check if honeypot field has been checked
    $formableForm = $message->getFormableForm();
    if ($formableForm->isHoneypotEnabled()) {
      $honeypotName = $formableForm->getHoneypotName();
      // prevent possible clashes with existing fields
      if ($message->hasField($honeypotName)) {
        $honeypotName = $formableForm->getHoneypotName(true);
      }
      if ($form_state->getValue($honeypotName)) {
        // TODO: message should be configurable (and translatable via config_translation module)
        $form_state->setErrorByName($honeypotName, $this->t('Robots are not welcome here.'));
      }
    }

    // Check if captcha field is correct
    if ($formableForm->isCaptchaEnabled()) {
      $captchaName = $formableForm->getCaptchaName();
      // prevent possible clashes with existing fields
      if ($message->hasField($captchaName)) {
        $captchaName = $formableForm->getCaptchaName(true);
      }
      $sentCaptchaValue = $form_state->getValue($captchaName);
      $expectedCaptchaValue = $formableForm->getCaptchaValue();
      $captchaIsWrong = !\is_string($sentCaptchaValue) || \strtolower(\trim($sentCaptchaValue)) !== \strtolower(\trim($expectedCaptchaValue));
      if ($captchaIsWrong) {
        // TODO: message should be configurable (and translatable via config_translation module)
        $form_state->setErrorByName($captchaName, $this->t('Please enter the correct word.'));
      }
    }

    // Check if flood control has been activated for sending emails.
    if (!$this->currentUser()->hasPermission('administer formable forms') && (!$this->currentUser()->hasPermission('administer users'))) {
      $limit = $this->config('formable.settings')->get('flood.limit');
      $interval = $this->config('formable.settings')->get('flood.interval');

      if (!$this->flood->isAllowed('formable', $limit, $interval)) {
        $form_state->setErrorByName('', $this->t('You cannot send more than %limit messages in @interval. Try again later.', [
          '%limit' => $limit,
          '@interval' => $this->dateFormatter->formatInterval($interval),
        ]));
      }
    }

    return $message;
  }

  /**
   * {@inheritdoc}
   */
  public function save(array $form, FormStateInterface $form_state) {
    $message = $this->entity;

    // Save the message.
    // TODO: not calling `$message->save()` should make
    // `FormableSqlContentEntityStorage::save()` overwrite obsolete
    $entityStorageResult = $message->save();
    $mailerStatusResult = $this->mailHandler->sendMailMessages($message);
    $formable_form = $message->getFormableForm();

    $this->flood->register('formable', $this->config('formable.settings')->get('flood.interval'));

    if ($entityStorageResult || $mailerStatusResult) {
      if ($submission_message = $formable_form->getMessage()) {
        $this->messenger()->addStatus($submission_message);
      }
    } else {
      $this->messenger()->addError($this->t('The message could not be sent.'));
    }

    // TODO: redirect to current page with anchor link to current form (when used as block)
    // To avoid false error messages caused by flood control, redirect away from
    // the form.
    $form_state->setRedirectUrl($formable_form->getRedirectUrl());
  }

}
