<?php

namespace Drupal\formable;

use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityListBuilder;
use Symfony\Component\HttpKernel\Exception\NotFoundHttpException;
use Drupal\Core\Entity\EntityTypeInterface;
use Drupal\Core\Entity\Query\QueryInterface;
use Drupal\formable\Entity\FormableForm;

/**
 * Defines a class to build a listing of message entities.
 *
 * TODO: needs tests
 *
 * @see \Drupal\formable\Entity\Message
 */
class MessageListBuilder extends EntityListBuilder {

  protected FormableForm $form;
  protected $formName;
  protected $fields;

  /**
   * Constructs a new EntityListBuilder object.
   *
   * @param \Drupal\Core\Entity\EntityTypeInterface $entity_type
   *   The entity type definition.
   * @param \Drupal\Core\Entity\EntityStorageInterface $storage
   *   The entity storage class.
   */
  public function __construct(
    EntityTypeInterface $entity_type,
    FormableSqlContentEntityStorage $storage
  ) {

    parent::__construct($entity_type, $storage);

    // TODO: How can I use dependency injection instead with url param?
    // - Probably not possible, because entity handlers are NOT services.
    // url param is specified in formable.routing.yml
    $formName = \Drupal::request()->attributes->get('_raw_variables')->get('formable_form');

    $form = \Drupal::entityTypeManager()->getStorage('formable_form')->load($formName);
    if (!$form) {
      throw new NotFoundHttpException();
    }

    $existingForms = \Drupal::service('entity_type.bundle.info')->getBundleInfo($this->entityTypeId);
    if (!isset($existingForms[$formName])) {
      throw new NotFoundHttpException();
    }

    $this->form = $form;
    $this->formName = $formName;
    $this->fields = \Drupal::service('entity_field.manager')->getFieldDefinitions('formable_message', $formName);
  }

  /**
   * Returns filtered list of form fields
   */
  public function getDisplayableFields() {
    return \array_filter($this->fields, function($field) {
      $ignoredFields = [
        'uuid',          // is used by rest, is a default for all entities
        'formable_form', // form ID
        'langcode',      // might be useful when filtering is implemented
        'ip_address',    // no relevance for daily usage
      ];
      return !\in_array($field->getName(), $ignoredFields);
    });
  }

  /**
   * {@inheritdoc}
   */
  public function buildHeader() {

    foreach ($this->getDisplayableFields() as $fieldName => $field) {
      $header[$fieldName] = $field->getLabel();
    }

    return $header + parent::buildHeader();
  }

  /**
   * {@inheritdoc}
   */
  public function buildRow(EntityInterface $entity) {

    $obfuscateMessagesForAdmin = $this->form->get('obfuscate_messages_for_admin');
    $isAdmin = \in_array('administrator', \Drupal::currentUser()->getRoles());

    foreach ($this->getDisplayableFields() as $fieldName => $field) {

      $value = $entity->get($fieldName)->value;

      if ('created' === $fieldName) {
        $dateFormatter = \Drupal::service('date.formatter');
        $value = $dateFormatter->format($value, 'short');
      }

      if ($obfuscateMessagesForAdmin && $isAdmin && \is_string($value)) {
        if (!\in_array($fieldName, ['created', 'referer'])) {
          $value = \str_rot13($value);
        }
      }

      // In case a form field has array values, flatten them
      if (\is_array($value)) {
        $value = \json_encode($value);
      }

      $row[$fieldName] = $value;
    }

    return $row + parent::buildRow($entity);
  }

  /**
   * Filter query by form/bundle and show newest first
   *
   * Returns a query object for loading entity IDs from the storage.
   *
   * @return \Drupal\Core\Entity\Query\QueryInterface
   *   A query object used to load entity IDs.
   */
  protected function getEntityListQuery(): QueryInterface {
    $query = parent::getEntityListQuery();
    $query->condition('formable_form', $this->formName);
    $query->sort($this->entityType->getKey('id'), 'DESC');

    return $query;
  }

}
