<?php

namespace Drupal\formable\Plugin\Block;

use Drupal\Core\Block\BlockBase;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Drupal\Core\Render\RendererInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Entity\EntityFormBuilderInterface;
use Drupal\Core\Session\AccountInterface;
use Drupal\Core\Access\AccessResult;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Block\BlockPluginInterface;

/**
 * Provides a 'Formable' block.
 *
 * Heavily inspired by the contact_block module
 * @see https://www.drupal.org/project/contact_block
 *
 * @Block(
 *   id = "formable_block",
 *   admin_label = @Translation("Form block"),
 *   category = @Translation("Forms"),
 * )
 */
class FormableBlock extends BlockBase implements ContainerFactoryPluginInterface {

  /**
   * The formable form configuration entity.
   *
   * @var \Drupal\formable\Entity\FormableForm
   */
  protected $formableForm;

  /**
   * The EntityTypeManager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManager
   */
  protected $entityTypeManager;

  /**
   * The EntityFormBuilder.
   *
   * @var \Drupal\Core\Entity\EntityFormBuilder
   */
  protected $entityFormBuilder;

  /**
   * The renderer.
   *
   * @var \Drupal\Core\Render\RendererInterface
   */
  protected $renderer;

  /**
   * Constructs a new SearchLocalTask.
   *
   * @param array $configuration
   *   A configuration array containing information about the plugin instance.
   * @param string $plugin_id
   *   The plugin ID for the plugin instance.
   * @param mixed $plugin_definition
   *   The plugin implementation definition.
   * @param \Drupal\Core\Render\RendererInterface $renderer
   *   The renderer.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity manager.
   * @param \Drupal\Core\Entity\EntityFormBuilderInterface $entity_form_builder
   *   The entity form builder.
   */
  public function __construct(
    array $configuration,
    $plugin_id,
    $plugin_definition,
    RendererInterface $renderer,
    EntityTypeManagerInterface $entity_type_manager,
    EntityFormBuilderInterface $entity_form_builder
  ) {
    parent::__construct($configuration, $plugin_id, $plugin_definition);
    $this->renderer = $renderer;
    $this->entityTypeManager = $entity_type_manager;
    $this->entityFormBuilder = $entity_form_builder;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static($configuration, $plugin_id, $plugin_definition,
      $container->get('renderer'),
      $container->get('entity_type.manager'),
      $container->get('entity.form_builder')
    );
  }

  /**
   * {@inheritdoc}
   */
  protected function blockAccess(AccountInterface $account) {
    $formableForm = $this->getFormableForm();

    // Deny access when the configured form has been deleted.
    if (empty($formableForm)) {
      return AccessResult::forbidden();
    }

    return $formableForm->access('block-view', $account, TRUE);
  }

  /**
   * {@inheritdoc}
   */
  public function build() {

    $formableForm = $this->getFormableForm();
    if (!$formableForm) return [];

    // If the form is invalid, show a warning, but still render it to be able
    // to preview it. Access control management prevents showing the form for
    // unprivileged users.
    if (!$formableForm->isValid()) {
      $this->messenger()->addWarning($this->t("The form @name is invalid because messages can't be stored and mails can't be sent. The form won't be displayed to anonymous users.", ['@name' => $formableForm->label()]));
    }

    $message = $this->entityTypeManager
      ->getStorage('formable_message')
      ->create([
        'formable_form' => $this->configuration['formable_form'],
      ]);

    $form = $this->entityFormBuilder->getForm($message);

    $form['#cache']['contexts'][] = 'user.permissions';
    $this->renderer->addCacheableDependency($form, $formableForm);

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function blockForm($form, FormStateInterface $form_state) {

    $options = $this->entityTypeManager
      ->getStorage('formable_form')
      ->loadMultiple();
    foreach ($options as $key => $option) {
      $options[$key] = $option->label();
    }

    // TODO: Add info text and link to config page if options are empty (no form exists)
    $form['formable_form'] = [
      '#type' => 'select',
      '#title' => $this->t('Form'),
      '#options' => $options,
      '#default_value' => $this->configuration['formable_form'] ?? null,
      '#required' => TRUE,
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function blockSubmit($form, FormStateInterface $form_state) {
    $this->setConfigurationValue('formable_form', $form_state->getValue('formable_form'));
  }

  /**
   * {@inheritdoc}
   */
  public function calculateDependencies() {

    $dependencies = array_merge_recursive(parent::calculateDependencies(), ['config' => []]);

    // Add the form as a dependency.
    if ($formableForm = $this->getFormableForm()) {
      $dependencies['config'][] = $formableForm->getConfigDependencyName();
    }

    return $dependencies;
  }

  /**
   * Loads the formable form entity.
   *
   * @return \Drupal\formable\Entity\FormableForm|null
   *   The formable form configuration entity. NULL if the entity does not exist.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  protected function getFormableForm() {
    if (!isset($this->formableForm)) {
      $formName = $this->configuration['formable_form'] ?? null;
      if ($formName) {
        $this->formableForm = $this->entityTypeManager
          ->getStorage('formable_form')
          ->load($formName);
      }
    }
    return $this->formableForm;
  }

}
